""" Raster data class managing I/O for multiple NetCDF files. """

import xarray as xr
import numpy as np
import pandas as pd
import netCDF4
from netCDF4 import MFDataset
from typing import Tuple

from geospade.crs import SpatialRef
from geospade.raster import Tile
from geospade.raster import MosaicGeometry

from veranda.utils import to_list
from veranda.raster.native.netcdf import NetCdf4File
from veranda.raster.mosaic.base import RasterDataReader, RasterDataWriter, RasterAccess


class NetCdfReader(RasterDataReader):
    """ Allows to read and manage a stack of NetCDF files. """
    def __init__(self, file_register, mosaic, stack_dimension='layer_id', stack_coords=None, tile_dimension='tile_id',
                 file_class=NetCdf4File, file_class_kwargs=None):
        """
        Constructor of `NetCdfReader`.

        Parameters
        ----------
        file_register : pd.Dataframe
            Data frame managing a stack/list of files containing the following columns:
                - 'filepath' : str
                    Full file path to a geospatial file.
                - 'layer_id' : object
                    Specifies an ID to which layer a file belongs to, e.g. a layer counter or a timestamp. Must
                    correspond to `stack_dimension`.
                - 'tile_id' : str
                    Tile name or ID to which tile a file belongs to.
        mosaic : geospade.raster.MosaicGeometry
            Mosaic representing the spatial allocation of the given files. The tiles of the mosaic have to match the
            ID's/names of the 'tile_id' column.
        stack_dimension : str, optional
            Dimension/column name of the dimension, where to stack the files along (first axis), e.g. time, bands etc.
            Defaults to 'layer_id', i.e. the layer ID's are used as the main coordinates to stack the files.
        stack_coords : list, optional
            Additional columns of `file_register` to use as coordinates. Defaults to None, i.e. only coordinates along
            `stack_dimension` are used.
        tile_dimension : str, optional
            Dimension/column name of the dimension containing tile ID's in correspondence with the tiles in `mosaic`.
            Defaults to 'tile_id'.
        file_class : class, optional
            Class to use for opening files. Defaults to `NetCdf4File`.
        file_class_kwargs : dict, optional
            Keyword arguments for calling `file_class`. Defaults to none.

        """
        super().__init__(file_register, mosaic, stack_dimension=stack_dimension, stack_coords=stack_coords,
                         tile_dimension=tile_dimension)

        file_class_kwargs = file_class_kwargs or dict()
        ref_filepath = self._file_register['filepath'].iloc[0]
        with file_class(ref_filepath, 'r', **file_class_kwargs) as nc_file:
            self._ref_data_variables = nc_file.data_variables
            self._ref_nodatavals = nc_file.nodatavals
            self._ref_scale_factors = nc_file.scale_factors
            self._ref_offsets = nc_file.offsets
            self._ref_dtypes = nc_file.dtypes
            self._ref_metadata = nc_file.metadata
            space_dims = nc_file.space_dims
            stack_dims = nc_file.stack_dims
            self._ref_space_dims = list(space_dims.keys()) if isinstance(space_dims, dict) else space_dims
            self._ref_stack_dims = list(stack_dims.keys()) if isinstance(stack_dims, dict) else stack_dims

    @classmethod
    def from_filepaths(cls, filepaths, mosaic_class=MosaicGeometry, mosaic_kwargs=None, tile_kwargs=None,
                       stack_dimension='layer_id', tile_dimension='tile_id',
                       file_class=NetCdf4File, file_class_kwargs=None,
                       **kwargs) -> "NetCdfReader":
        """
        Creates a `NetCdfReader` instance as one stack of NetCDF files.

        Parameters
        ----------
        filepaths : list of str
            List of full system paths to a NetCDF file.
        mosaic_class : geospade.raster.MosaicGeometry, optional
            Mosaic class used to manage the spatial properties of the file stack. If None, the most generic mosaic will
            be used by default. The initialised mosaic will only contain one tile.
        mosaic_kwargs : dict, optional
            Additional arguments for initialising `mosaic_class`.
        tile_kwargs : dict, optional
            Additional arguments for initialising a tile class associated with `mosaic_class`.
        stack_dimension : str, optional
            Dimension/column name of the dimension, where to stack the files along (first axis), e.g. time, bands etc.
            Defaults to 'layer_id', i.e. the layer ID's are used as the main coordinates to stack the files.
        tile_dimension : str, optional
            Dimension/column name of the dimension containing tile ID's in correspondence with the tiles in `mosaic`.
            Defaults to 'tile_id'.
        file_class : class, optional
            Class to use for opening files. Defaults to `NetCdf4File`.
        file_class_kwargs : dict, optional
            Keyword arguments for calling `file_class`. Defaults to none.
        kwargs : dict, optional
            Key-word arguments for the `NetCdfReader` constructor.

        Returns
        -------
        NetCdfReader

        """
        mosaic_kwargs = mosaic_kwargs or dict()
        tile_kwargs = tile_kwargs or dict()
        file_class_kwargs = file_class_kwargs or dict()

        n_filepaths = len(filepaths)
        file_register_dict = dict()
        file_register_dict['filepath'] = filepaths
        file_register_dict[tile_dimension] = ['0'] * n_filepaths
        file_register_dict[stack_dimension] = list(range(n_filepaths))
        file_register = pd.DataFrame(file_register_dict)

        ref_filepath = filepaths[0]
        with file_class(ref_filepath, 'r', **file_class_kwargs) as nc_file:
            sref_wkt = nc_file.sref_wkt
            geotrans = nc_file.geotrans
            n_rows, n_cols = nc_file.raster_shape

        tile_class = mosaic_class.get_tile_class()
        tile = tile_class(n_rows, n_cols, sref=SpatialRef(sref_wkt), geotrans=geotrans, name='0', **tile_kwargs)
        mosaic_geom = mosaic_class.from_tile_list([tile], check_consistency=False, **mosaic_kwargs)

        return cls(file_register, mosaic_geom, stack_dimension=stack_dimension, tile_dimension=tile_dimension,
                   **kwargs)

    @classmethod
    def from_mosaic_filepaths(cls, filepaths, mosaic_class=MosaicGeometry, mosaic_kwargs=None,
                              stack_dimension='layer_id', tile_dimension='tile_id',
                              file_class=NetCdf4File, file_class_kwargs=None, **kwargs) -> "NetCdfReader":
        """
        Creates a `NetCdfReader` instance as multiple stacks of NetCDF files.

        Parameters
        ----------
        filepaths : list of str
            List of full system paths to a NetCDF file.
        mosaic_class : geospade.raster.MosaicGeometry, optional
            Mosaic class used to manage the spatial properties of the file stacks. If None, the most generic mosaic
            will be used by default.
        mosaic_kwargs : dict, optional
            Additional arguments for initialising `mosaic_class`.
        stack_dimension : str, optional
            Dimension/column name of the dimension, where to stack the files along (first axis), e.g. time, bands etc.
            Defaults to 'layer_id', i.e. the layer ID's are used as the main coordinates to stack the files.
        tile_dimension : str, optional
            Dimension/column name of the dimension containing tile ID's in correspondence with the tiles in `mosaic`.
            Defaults to 'tile_id'.
        file_class : class, optional
            Class to use for opening files. Defaults to `NetCdf4File`.
        file_class_kwargs : dict, optional
            Keyword arguments for calling `file_class`. Defaults to none.
        kwargs : dict, optional
            Key-word arguments for the `NetCdfReader` constructor.

        Returns
        -------
        NetCdfReader

        """
        mosaic_kwargs = mosaic_kwargs or dict()
        file_class_kwargs = file_class_kwargs or dict()
        file_register_dict = dict()
        file_register_dict['filepath'] = filepaths
        tile_class = mosaic_class.get_tile_class()
        tiles, tile_ids, layer_ids = RasterDataReader._create_tile_and_layer_info_from_files(filepaths, tile_class,
                                                                                             file_class,
                                                                                             file_class_kwargs)

        file_register_dict[tile_dimension] = tile_ids
        file_register_dict[stack_dimension] = layer_ids
        file_register = pd.DataFrame(file_register_dict)

        mosaic_geom = mosaic_class.from_tile_list(tiles, check_consistency=False, **mosaic_kwargs)

        return cls(file_register, mosaic_geom, stack_dimension=stack_dimension, tile_dimension=tile_dimension,
                   **kwargs)

    def apply_nan(self):
        """ Converts no data values given as an attribute '_FillValue' or keyword `nodatavals` to np.nan. """
        super().apply_nan(nodatavals=self._ref_nodatavals)

    def read(self, data_variables=None, engine='netcdf4', agg_dim='time', parallel=True, compute=True, auto_decode=False,
             decoder=None, decoder_kwargs=None, **kwargs) -> "NetCdfReader":
        """
        Reads NetCdf data from disk and assigns it to the class.

        Parameters
        ----------
        data_variables : list, optional
            Data variables to read. Default is to read all available data variables.
        engine : str, optional
            Engine used in the background to read NetCDF data. The following options are available:
                - 'netcdf4' : Uses the netCDF4 library to create an `MFDataset` object.
                - 'xarray' : Uses xarray's `open_mfdataset` function.
        agg_dim : str, optional
            Dimension to aggregate on (defaults to 'layer_id').
        parallel : bool, optional
            Flag to activate parallelisation or not when using 'xarray' as an engine. Defaults to True.
        compute : bool, optional
            True if values from a dask array should be loaded into RAM (default).
        auto_decode : bool, optional
            True if NetCDF data should be decoded according to the information available in its metadata. Defaults to
            False.
        decoder : callable, optional
            Function allowing to decode NetCDF data read from disk.
        decoder_kwargs : dict, optional
            Keyword arguments for the decoder.

        """
        data_variables = to_list(data_variables)
        dst_tile = Tile.from_extent(self._mosaic.outer_extent, sref=self._mosaic.sref,
                                    x_pixel_size=self._mosaic.x_pixel_size,
                                    y_pixel_size=self._mosaic.y_pixel_size,
                                    name='0')
        if engine == 'netcdf4':
            data = self.__read_netcdf4(dst_tile, data_variables=data_variables, agg_dim=agg_dim,
                                       auto_decode=auto_decode, decoder=decoder, decoder_kwargs=decoder_kwargs,
                                       **kwargs)
        elif engine == 'xarray':
            data = self.__read_xarray(dst_tile, data_variables=data_variables, parallel=parallel,
                                      agg_dim=agg_dim, compute=compute, auto_decode=auto_decode, decoder=decoder,
                                      decoder_kwargs=decoder_kwargs, **kwargs)
        else:
            err_msg = f"Engine '{engine}' is not supported!"
            raise ValueError(err_msg)

        self._data_geom = dst_tile
        self._data = data
        self._add_grid_mapping()
        return self

    def __load_data_per_data_variable(self, ds, data_variable, raster_access,
                                      decoder=None, decoder_kwargs=None) -> xr.DataArray:
        """
        Selects and slices an xarray data array from an xarray dataset.

        Parameters
        ----------
        ds : xr.Dataset
            Dataset to subset.
        data_variable : str
            Data variable to select.
        raster_access : RasterAccess
            Helper instance to slice the data array.
        decoder : callable, optional
            Function allowing to decode NetCDF data read from disk.
        decoder_kwargs : dict, optional
            Keyword arguments for the decoder.

        Returns
        -------
        dar : xr.DataArray
            Decoded data subset.

        """
        dar = ds[data_variable][..., raster_access.src_row_slice, raster_access.src_col_slice]
        if decoder:
            dar = decoder(dar, nodataval=self._ref_nodatavals[data_variable],
                          data_variable=data_variable,
                          scale_factor=self._ref_scale_factors[data_variable],
                          offset=self._ref_offsets[data_variable],
                          dtype=self._ref_dtypes[data_variable],
                          **decoder_kwargs)
        return dar

    def __read_netcdf4(self, dst_tile, data_variables=None, agg_dim='layer_id', auto_decode=False,
                       decoder=None, decoder_kwargs=None, **kwargs) -> xr.Dataset:
        """
        Reads NetCDF data using the `MFDataset` class of the netCDF4 library.

        Parameters
        ----------
        dst_tile : geospade.raster.Tile
            Target tile representing the spatial extent of the data window to read from.
        data_variables : list, optional
            Data variables to read. Default is to read all available data variables.
        agg_dim : str, optional
            Dimension to aggregate on (defaults to 'layer_id').
        auto_decode : bool, optional
            True if NetCDF data should be decoded according to the information available in its metadata. Defaults to
            False.
        decoder : callable, optional
            Function allowing to decode NetCDF data read from disk.
        decoder_kwargs : dict, optional
            Keyword arguments for the decoder.

        Returns
        -------
        xr.Dataset :
            Read NetCDF variables represented as an xarray.Dataset instance.

        """
        data_variables = data_variables or self._ref_data_variables
        decoder_kwargs = decoder_kwargs or dict()
        data = [self.__load_data_per_tile_netcdf4(src_tile, dst_tile, data_variables, agg_dim, auto_decode, decoder,
                                                  decoder_kwargs)
                for src_tile in self._mosaic.tiles]

        return xr.combine_by_coords(data)

    def __post_proc_data_netcdf4(self, ar, dw_tile, dm_tile, nodataval=0) -> xr.DataArray:
        """
        Masks the given array.

        Parameters
        ----------
        ar : np.ndarray or xr.DataArray
            Array to mask.
        dw_tile : Tile
            Tile representing the data window to read from.
        dm_tile : Tile
            Tile to extract a data mask from.
        nodataval : float, optional
            No data value being assigned where the mask values evaluate to false (defaults to 0).

        Returns
        -------
        ar : np.ndarray or xr.DataArray
            Masked array.

        """
        if dm_tile.mask is not None:
            raster_access = RasterAccess(dm_tile, dw_tile, src_root_raster_geom=dw_tile)
            tile_mask = np.zeros(dw_tile.shape)
            tile_mask[raster_access.src_row_slice, raster_access.src_col_slice] = dm_tile.mask
            ar[:, ~tile_mask.astype(bool)] = nodataval

        return ar

    def __load_data_per_data_variable_netcdf4(self, ds, data_variable, dw_tile, dm_tile, raster_access,
                                              decoder=None, decoder_kwargs=None) -> xr.DataArray:
        """
        Selects, mask and slices a netCDF4 data variable.

        Parameters
        ----------
        ds : xr.Dataset
            Dataset to subset.
        data_variable : str
            Data variable to select.
        dw_tile : Tile
            Tile representing the data window to read from.
        dm_tile : Tile
            Tile to extract a data mask from.
        raster_access : RasterAccess
            Helper instance to slice the data array.
        decoder : callable, optional
            Function allowing to decode NetCDF data read from disk.
        decoder_kwargs : dict, optional
            Keyword arguments for the decoder.

        Returns
        -------
        dar : xr.DataArray
            Decoded data subset.

        """
        dar = self.__load_data_per_data_variable(ds, data_variable, raster_access,
                                                 decoder, decoder_kwargs)
        dar = self.__post_proc_data_netcdf4(dar, dw_tile, dm_tile, self._ref_nodatavals[data_variable])

        return dar

    def __load_data_per_tile_netcdf4(self, src_tile, dst_tile, data_variables, agg_dim='layer_id', auto_decode=False,
                                     decoder=None, decoder_kwargs=None) -> xr.Dataset:
        """
        Creates an xarray dataset per tile for a given set of data variables from a multi-file netCDF4 dataset.

        Parameters
        ----------
        src_tile : Tile
            Source tile representing the spatial extent of the data window to read from.
        dst_tile : Tile
            Target tile representing the spatial extent of the data window to write to.
        data_variables : list
            Data variables to read.
        agg_dim : str, optional
            Dimension to aggregate on (defaults to 'layer_id').
        auto_decode : bool, optional
            True if NetCDF data should be decoded according to the information available in its metadata. Defaults to
            False.
        decoder : callable, optional
            Function allowing to decode NetCDF data read from disk.
        decoder_kwargs : dict, optional
            Keyword arguments for the decoder.

        Returns
        -------
        xr.Dataset:
            Dataset corresponding to the spatial and variable selection.

        """
        tile_id = src_tile.parent_root.name
        # create new tile representing the part to actually read
        dw_tile = src_tile.parent_root.slice_by_geom(dst_tile)
        raster_access = RasterAccess(dw_tile, dst_tile)
        file_register = self.file_register.loc[self.file_register[self._tile_dim] == tile_id]
        filepaths = list(file_register['filepath'])
        nc_ds = MFDataset(filepaths, aggdim=agg_dim)
        nc_ds.set_auto_maskandscale(auto_decode)
        tile_data = {data_variable: self.__load_data_per_data_variable_netcdf4(nc_ds, data_variable, dw_tile, src_tile,
                                                                               raster_access,
                                                                               decoder, decoder_kwargs)
                     for data_variable in data_variables}
        metadata = {data_variable: NetCdf4File.get_metadata(nc_ds[data_variable])
                    for data_variable in data_variables}

        times = netCDF4.num2date(nc_ds['time'][:],  # TODO: generalise stack dimension
                                 units=getattr(nc_ds['time'], 'units', None),
                                 calendar=getattr(nc_ds['time'], 'calendar', 'standard'),
                                 only_use_cftime_datetimes=False,
                                 only_use_python_datetimes=True)

        return self._to_xarray(tile_data, dw_tile, times, metadata)

    def __read_xarray(self, dst_tile, data_variables=None, parallel=True, agg_dim='layer_id', compute=True,
                      auto_decode=False, decoder=None, decoder_kwargs=None, **kwargs) -> xr.Dataset:
        """
        Reads NetCDF data using the `open_mfdataset` function of the xarray library.

        Parameters
        ----------
        dst_tile : geospade.raster.Tile
            Target tile representing the spatial extent of the data window to read from.
        data_variables : list, optional
            Data variables to read. Default is to read all available data variables.
        parallel : bool, optional
            Flag to activate parallelisation or not when using 'xarray' as an engine. Defaults to True.
        agg_dim : str, optional
            Dimension to aggregate on (defaults to 'layer_id').
        compute : bool, optional
            True if values from a dask array should be loaded into RAM (default).
        auto_decode : bool, optional
            True if NetCDF data should be decoded according to the information available in its metadata. Defaults to
            False.
        decoder : callable, optional
            Function allowing to decode NetCDF data read from disk.
        decoder_kwargs : dict, optional
            Keyword arguments for the decoder.

        Returns
        -------
        xr.Dataset :
            Read NetCDF variables represented as an xarray.Dataset instance.

        """
        data_variables = data_variables or self._ref_data_variables
        decoder_kwargs = decoder_kwargs or dict()
        data = [self.__load_data_per_tile_xarray(src_tile, dst_tile, data_variables, parallel, agg_dim, compute,
                                                 auto_decode, decoder, decoder_kwargs, **kwargs)
                for src_tile in self._mosaic.tiles]

        return xr.combine_by_coords(data)

    def __post_proc_data_xarray(self, dar, dw_tile, dm_tile, nodataval=0, compute=True) -> xr.DataArray:
        """
        Masks the given array.

        Parameters
        ----------
        dar : xr.DataArray
            Array to mask.
        dw_tile : Tile
            Tile representing the data window to read from.
        dm_tile : Tile
            Tile to extract a data mask from.
        nodataval : float, optional
            No data value being assigned where the mask values evaluate to false (defaults to 0).
        compute : bool, optional
            True if values from a dask array should be loaded into RAM (default).

        Returns
        -------
        ar : xr.DataArray
            Masked array.

        """
        if compute:
            dar = dar.compute()
        if dm_tile.mask is not None:
            raster_access = RasterAccess(dm_tile, dw_tile, src_root_raster_geom=dw_tile)
            tile_mask = np.zeros(dw_tile.shape)
            tile_mask[raster_access.src_row_slice, raster_access.src_col_slice] = dm_tile.mask
            dar = dar.where(tile_mask.astype(bool), nodataval)

        return dar

    def __load_data_per_data_variable_xarray(self, ds, data_variable, dw_tile, dm_tile, raster_access,
                                      decoder=None, decoder_kwargs=None, compute=True) -> xr.DataArray:
        """
        Selects, mask and slices an xarray data array.

        Parameters
        ----------
        ds : xr.Dataset
            Dataset to subset.
        data_variable : str
            Data variable to select.
        dw_tile : Tile
            Tile representing the data window to read from.
        dm_tile : Tile
            Tile to extract a data mask from.
        raster_access : RasterAccess
            Helper instance to slice the data array.
        decoder : callable, optional
            Function allowing to decode NetCDF data read from disk.
        decoder_kwargs : dict, optional
            Keyword arguments for the decoder.
        compute : bool, optional
            True if values from a dask array should be loaded into RAM (default).

        Returns
        -------
        dar : xr.DataArray
            Decoded data subset.

        """
        dar = self.__load_data_per_data_variable(ds, data_variable, raster_access, decoder, decoder_kwargs)
        dar = self.__post_proc_data_xarray(dar, dw_tile, dm_tile, self._ref_nodatavals[data_variable], compute)

        return dar

    def __load_data_per_tile_xarray(self, src_tile, dst_tile, data_variables, parallel=True,
                                    agg_dim='layer_id', compute=True, auto_decode=False, decoder=None,
                                    decoder_kwargs=None, **kwargs) -> xr.Dataset:
        """
        Creates an xarray dataset per tile for a given set of data variables.

        Parameters
        ----------
        src_tile : Tile
            Source tile representing the spatial extent of the data window to read from.
        dst_tile : Tile
            Target tile representing the spatial extent of the data window to write to.
        data_variables : list
            Data variables to read.
        parallel : bool, optional
            Flag to activate parallelisation or not when using 'xarray' as an engine. Defaults to True.
        agg_dim : str, optional
            Dimension to aggregate on (defaults to 'layer_id').
        compute : bool, optional
            True if values from a dask array should be loaded into RAM (default).
        auto_decode : bool, optional
            True if NetCDF data should be decoded according to the information available in its metadata. Defaults to
            False.
        decoder : callable, optional
            Function allowing to decode NetCDF data read from disk.
        decoder_kwargs : dict, optional
            Keyword arguments for the decoder.

        Returns
        -------
        xr.Dataset:
            Dataset corresponding to the spatial and variable selection.

        """
        tile_id = src_tile.parent_root.name
        # create new tile representing the part to actually read
        dw_tile = src_tile.parent_root.slice_by_geom(dst_tile)
        raster_access = RasterAccess(dw_tile, dst_tile)
        file_register = self.file_register.loc[self.file_register[self._tile_dim] == tile_id]
        filepaths = file_register['filepath']
        xr_ds = xr.open_mfdataset(filepaths, concat_dim=agg_dim, combine="nested", data_vars='minimal',
                                  coords='minimal', compat='override', parallel=parallel,
                                  mask_and_scale=auto_decode, **kwargs)
        data_tile = {data_variable: self.__load_data_per_data_variable_xarray(xr_ds, data_variable, dw_tile, src_tile,
                                                                              raster_access, decoder,
                                                                              decoder_kwargs, compute)
                     for data_variable in data_variables}
        ref_coords = data_tile[data_variables[0]].coords
        return xr.Dataset(data_tile, coords=ref_coords, attrs=xr_ds.attrs)

    def _to_xarray(self, data, tile, times, metadata) -> xr.Dataset:
        """
        Converts NetCDF data being available as a NumPy array to an xarray dataset.

        Parameters
        ----------
        data : dict
            Dictionary mapping data variables with NetCDF variable data being available as a NumPy array.
        tile : geospade.raster.Tile
            Tile representing the spatial extent of `data`.
        times : list
            List of datetime instances representing the temporal coordinates of the image stack.
        metadata : dict
            Metadata attributes for each data variable.

        Returns
        -------
        xrds : xr.Dataset

        """
        space_dim_names = self._ref_space_dims
        stack_dim_names = self._ref_stack_dims
        all_dim_names = stack_dim_names + space_dim_names
        coord_dict = dict()
        coord_dict[stack_dim_names[0]] = times
        coord_dict[space_dim_names[0]] = tile.y_coords
        coord_dict[space_dim_names[1]] = tile.x_coords

        data_variables = list(data.keys())
        xar_dict = {data_variable: xr.DataArray(data[data_variable], coords=coord_dict, dims=all_dim_names,
                                                attrs=metadata[data_variable])
                    for data_variable in data_variables}
        xrds = xr.Dataset(data_vars=xar_dict)

        return xrds


class NetCdfWriter(RasterDataWriter):
    """ Allows to write and manage a stack of NetCDF files. """
    def __init__(self, mosaic, file_register=None, data=None, stack_dimension='layer_id', stack_coords=None,
                 tile_dimension='tile_id', dirpath=None, fn_pattern='{layer_id}.tif', fn_formatter=None):
        """
        Constructor of `NetCdfWriter`.

        Parameters
        ----------
        mosaic : geospade.raster.MosaicGeometry
            Mosaic representing the spatial allocation of the given files. The tiles of the mosaic have to match the
            ID's/names of the 'tile_id' column.
        file_register : pd.Dataframe, optional
            Data frame managing a stack/list of files containing the following columns:
                - 'filepath' : str
                    Full file path to a geospatial file.
                - 'layer_id' : object
                    Specifies an ID to which layer a file belongs to, e.g. a layer counter or a timestamp. Must
                    correspond to `stack_dimension`.
                - 'tile_id' : str
                    Tile name or ID to which tile a file belongs to.
            If it is None, then the constructor tries to create a file from other keyword arguments, i.e. `data`,
            `dirpath`, `fn_pattern`, and `fn_formatter`.
        data : xr.Dataset, optional
            Raster data stored in memory. It must match the spatial sampling and CRS of the mosaic, but not its spatial
            extent or tiling. Moreover, the dimension of the mosaic along the first dimension (stack/file dimension),
            must match the entries/filepaths in `file_register`.
        stack_dimension : str, optional
            Dimension/column name of the dimension, where to stack the files along (first axis), e.g. time, bands etc.
            Defaults to 'layer_id', i.e. the layer ID's are used as the main coordinates to stack the files.
        stack_coords : list, optional
            Additional columns of `file_register` to use as coordinates. Defaults to None, i.e. only coordinates along
            `stack_dimension` are used.
        tile_dimension : str, optional
            Dimension/column name of the dimension containing tile ID's in correspondence with the tiles in `mosaic`.
            Defaults to 'tile_id'.
        dirpath : str, optional
            Directory path to the folder where the NetCDF files should be written to. Defaults to None, i.e. the
            current working directory is used.
        fn_pattern : str, optional
            Pattern for the filename of the new NetCDF files. To fill in specific parts of the new file name with
            information from the file register, you can specify the respective file register column names in curly
            brackets and add them to the pattern string as desired. Defaults to '{layer_id}.tif'.
        fn_formatter : dict, optional
            Dictionary mapping file register column names with functions allowing to encode their values as strings.

        """

        super().__init__(mosaic, file_register=file_register, data=data, stack_dimension=stack_dimension,
                         stack_coords=stack_coords, tile_dimension=tile_dimension, dirpath=dirpath,
                         fn_pattern=fn_pattern, fn_formatter=fn_formatter)

    @classmethod
    def from_data(self, data, filepath, mosaic=None, tile_dimension='tile_id', **kwargs) -> "NetCdfWriter":
        """
        Creates `NetCdfWriter` instance from an xarray.Dataset instance and a target file path, i.e. this function
        should help to write/export the whole image stack to one file.

        Parameters
        ----------
        data : xr.Dataset
            Dataset to write to disk.
        filepath : str
            Full system path to NetCDF file to write to.
        mosaic : geospade.raster.MosaicGeometry, optional
            Mosaic representing the spatial allocation of the given file. The tiles of the mosaic have to match the
            ID's/names of the 'tile_id' column. If it is None, a one-tile mosaic will be created from the given
            mosaic.
        tile_dimension : str, optional
            Dimension/column name of the dimension containing tile ID's in correspondence with the tiles in `mosaic`.
            Defaults to 'tile_id'.
        kwargs : dict, optional
            Key-word arguments for initialising the `NetCdfWriter` class.

        Returns
        -------
        NetCdfWriter

        """
        file_register_dict = dict()
        file_register_dict[tile_dimension] = ['0']
        file_register_dict['filepath'] = [filepath]
        file_register = pd.DataFrame(file_register_dict)
        return super().from_xarray(data, file_register, mosaic=mosaic, tile_dimension=tile_dimension, **kwargs)

    def __get_encoding_info_from_data(self, data, data_variables) -> Tuple[dict, dict, dict, dict]:
        """
        Extracts encoding information from an xarray dataset.

        Parameters
        ----------
        data : xr.Dataset
            Data to extract encoding info from.
        data_variables : list of str
            List of data variable names.

        Returns
        -------
        nodatavals : dict
            Band number mapped to no data value (defaults to 0).
        scale_factors : dict
            Band number mapped to scale factor (defaults to 1).
        offsets : dict
            Band number mapped to offset (defaults to 0).
        dtypes : dict
            Band number mapped to data type.

        """
        nodatavals = dict()
        scale_factors = dict()
        offsets = dict()
        dtypes = dict()
        for data_variable in data_variables:
            dtypes[data_variable] = data[data_variable].data.dtype.name
            nodatavals[data_variable] = data[data_variable].attrs.get('_FillValue', 0)
            scale_factors[data_variable] = data[data_variable].attrs.get('scale_factor', 1)
            offsets[data_variable] = data[data_variable].attrs.get('add_offset', 0)

        return nodatavals, scale_factors, offsets, dtypes

    def write(self, data, use_mosaic=False, data_variables=None, encoder=None, encoder_kwargs=None, overwrite=False,
              unlimited_dims=None, **kwargs):
        """
        Writes a certain chunk of NetCDF data to disk.

        Parameters
        ----------
        data : xr.Dataset
            Data chunk to be written to disk or being appended to existing mosaic.
        use_mosaic : bool, optional
            True if data should be written according to the mosaic.
            False if data composes a new tile and should not be tiled (default).
        data_variables : list of str, optional
            Data variables to write. Defaults to None, i.e. all data variables are written.
        encoder : callable, optional
            Function allowing to encode a xarray dataset before writing it to disk.
        encoder_kwargs : dict, optional
            Keyword arguments for the encoder.
        overwrite : bool, optional
            True if the NetCDF file(s) should be overwritten, False if not (default).
        unlimited_dims : list of str or str, optional
            List of dimension names specifying the dimensions to be stored as unlimited.
        kwargs : dict, optional
            Key-word arguments for creating a `NetCdf4File` instance.

        """
        data_geom = self.raster_geom_from_data(data, sref=self.mosaic.sref)
        unlimited_dims = to_list(unlimited_dims)
        data_filt = data if data_variables is None else data[data_variables]
        data_variables = list(data_filt.data_vars)
        all_dims = list(data_filt.dims)
        space_dims = all_dims[-2:]
        stack_dim_names = all_dims[:-2]
        nodatavals, scale_factors, offsets, dtypes = self.__get_encoding_info_from_data(data, data_variables)

        for filepath, file_group in self._file_register.groupby('filepath'):
            tile_id = file_group.iloc[0].get(self._tile_dim, '0')

            if use_mosaic:
                src_tile = self._mosaic[tile_id]
                if not src_tile.intersects(data_geom):
                    continue
                dst_tile = data_geom.slice_by_geom(src_tile, inplace=False)
                data_write = data_filt.sel(**{space_dims[0]: dst_tile.y_coords, space_dims[1]: dst_tile.x_coords})
            else:
                dst_tile = data_geom
                src_tile = data_geom
                data_write = data_filt

            file_coords = list(file_group[self._file_dim])
            data_write = data_write.sel(**{self._file_dim: file_coords})
            data_write = data_write[data_variables]
            stack_dims = {stack_dim_name: None if stack_dim_name in unlimited_dims else len(data_write[stack_dim_name])
                          for stack_dim_name in stack_dim_names}

            file_id = file_group.iloc[0].get('file_id', None)
            if file_id is None:
                gt_driver = NetCdf4File(filepath, mode='w', geotrans=src_tile.geotrans, sref_wkt=src_tile.sref.wkt,
                                        stack_dims=stack_dims,
                                        space_dims={space_dims[0]: src_tile.n_rows, space_dims[1]: src_tile.n_cols},
                                        data_variables=data_variables, dtypes=dtypes,
                                        scale_factors=scale_factors, offsets=offsets, nodatavals=nodatavals,
                                        attrs={'time': {'units': 'days since 1950-01-01 00:00:00'}},  # TODO: make this more flexible (this needs to be defined from outside)
                                        metadata=data_write.attrs, **kwargs)
                file_id = len(list(self._files.keys())) + 1
                self._files[file_id] = gt_driver
                self._file_register.loc[file_group.index, 'file_id'] = file_id

            nc_file = self._files[file_id]
            raster_access = RasterAccess(dst_tile, src_tile, src_root_raster_geom=data_geom)
            nc_file.write(data_write, row=raster_access.dst_window[0], col=raster_access.dst_window[1],
                          encoder=encoder, encoder_kwargs=encoder_kwargs)

    def export(self, use_mosaic=False, data_variables=None, encoder=None, encoder_kwargs=None, overwrite=False,
               unlimited_dims=None, **kwargs):
        """
        Writes all internally stored data to disk.

        Parameters
        ----------
        use_mosaic : bool, optional
            True if data should be written according to the mosaic.
            False if data composes a new tile and should not be tiled (default).
        data_variables : list of str, optional
            Data variables to write. Defaults to None, i.e. all data variables are written.
        encoder : callable, optional
            Function allowing to encode an xarray dataset before writing it to disk.
        encoder_kwargs : dict, optional
            Keyword arguments for the encoder.
        overwrite : bool, optional
            True if the NetCDF file(s) should be overwritten, False if not (default).
        unlimited_dims : list of str or str, optional
            List of dimension names specifying the dimensions to be stored as unlimited.
        kwargs : dict, optional
            Key-word arguments for creating a `NetCdf4File` instance.

        """
        self.write(self.data_view, use_mosaic, data_variables, encoder, encoder_kwargs, overwrite, unlimited_dims,
                   **kwargs)


if __name__ == '__main__':
    pass



""" Raster data class managing I/O for multiple GeoTIFF files. """

import os
import uuid
import tempfile
import xarray as xr
import numpy as np
import pandas as pd
from osgeo import gdal
from datetime import datetime
from typing import Tuple, List
from multiprocessing import Pool, RawArray

from geospade.crs import SpatialRef
from geospade.raster import Tile
from geospade.raster import MosaicGeometry

from veranda.utils import to_list
from veranda.raster.native.geotiff import GeoTiffFile
from veranda.raster.native.geotiff import create_vrt_file
from veranda.raster.gdalport import GDAL_TO_NUMPY_DTYPE
from veranda.raster.mosaic.base import RasterDataReader, RasterDataWriter, RasterAccess

PROC_OBJS = {}


def read_init(fr, am, sm, sd, td, ad, dc, dk):
    """ Helper method for setting the entries of global variable `PROC_OBJS` to be available during multiprocessing. """
    PROC_OBJS['global_file_register'] = fr
    PROC_OBJS['access_map'] = am
    PROC_OBJS['shm_map'] = sm
    PROC_OBJS['stack_dimension'] = sd
    PROC_OBJS['tile_dimension'] = td
    PROC_OBJS['auto_decode'] = ad
    PROC_OBJS['decoder'] = dc
    PROC_OBJS['decoder_kwargs'] = dk


class GeoTiffAccess(RasterAccess):
    """
    Helper class to build the link between indexes of the source array (access) and the target array (assignment).
    The base class `RasterAccess` is extended with some properties needed for accessing GeoTIFF files.

    """
    def __init__(self, src_raster_geom, dst_raster_geom, src_root_raster_geom=None):
        """
        Constructor of `GeoTiffAccess`.

        Parameters
        ----------
        src_raster_geom : geospade.raster.RasterGeometry
            Raster geometry representing the extent and indices of the array to access.
        dst_raster_geom : geospade.raster.RasterGeometry
            Raster geometry representing the extent and indices of the array to assign.
        src_root_raster_geom : geospade.raster.RasterGeometry, optional
            Raster geometry representing the origin to which `src_raster_geom` should be referred to. Defaults to None,
            i.e. the root parent of `src_raster_geom` is used.

        """
        super().__init__(src_raster_geom, dst_raster_geom, src_root_raster_geom=src_root_raster_geom)
        self.src_wkt = src_raster_geom.parent_root.sref.wkt
        self.src_geotrans = src_raster_geom.parent_root.geotrans
        self.src_shape = src_raster_geom.parent_root.shape

    @property
    def gdal_args(self) -> Tuple[int, int, int, int]:
        """ Prepares the needed positional arguments for GDAL's `ReadAsArray()` function. """
        min_row, min_col, max_row, max_col = self.src_window
        n_cols, n_rows = max_col - min_col + 1, max_row - min_row + 1
        return min_col, min_row, n_cols, n_rows

    @property
    def read_args(self) -> Tuple[int, int, int, int]:
        """ Prepares the needed positional arguments for the `read()` function of the internal GeoTIFF native. """
        min_col, min_row, n_cols, n_rows = self.gdal_args
        return min_row, min_col, n_rows, n_cols


class GeoTiffReader(RasterDataReader):
    """ Allows to read and manage a stack of GeoTIFF data. """
    def __init__(self, file_register, mosaic, stack_dimension='layer_id', stack_coords=None, tile_dimension='tile_id',
                 space_dims=None, file_class=GeoTiffFile, file_class_kwargs=None):
        """
        Constructor of `GeoTiffReader`.

        Parameters
        ----------
        file_register : pd.Dataframe, optional
            Data frame managing a stack/list of files containing the following columns:
                - 'filepath' : str
                    Full file path to a geospatial file.
                - 'layer_id' : object
                    Specifies an ID to which layer a file belongs to, e.g. a layer counter or a timestamp. Must
                    correspond to `stack_dimension`.
                - 'tile_id' : str
                    Tile name or ID to which tile a file belongs to.
        mosaic : geospade.raster.MosaicGeometry
            Mosaic representing the spatial allocation of the given files. The tiles of the mosaic have to match the
            ID's/names of the 'tile_id' column.
        stack_dimension : str, optional
            Dimension/column name of the dimension, where to stack the files along (first axis), e.g. time, bands etc.
            Defaults to 'layer_id', i.e. the layer ID's are used as the main coordinates to stack the files.
        stack_coords : list, optional
            Additional columns of `file_register` to use as coordinates. Defaults to None, i.e. only coordinates along
            `stack_dimension` are used.
        tile_dimension : str, optional
            Dimension/column name of the dimension containing tile ID's in correspondence with the tiles in `mosaic`.
            Defaults to 'tile_id'.
        space_dims : list, optional
            Dictionary containing the spatial dimension names. By default it is set to ['y', 'x'].
        file_class : class, optional
            Class for opening GeoTIFF files. Defaults to `GeoTiffFile`.
        file_class_kwargs : dict, optional
            Keyword arguments for `file_class`.

        """
        super().__init__(file_register, mosaic, stack_dimension=stack_dimension, stack_coords=stack_coords,
                         tile_dimension=tile_dimension)

        file_class_kwargs = file_class_kwargs or dict()
        ref_filepath = self._file_register['filepath'].iloc[0]
        with file_class(ref_filepath, 'r', **file_class_kwargs) as gt_file:
            self._ref_dtypes = gt_file.dtypes
            self._ref_nodatavals = gt_file.nodatavals

        self._ref_space_dims = space_dims or ['y', 'x']

    @classmethod
    def from_filepaths(cls, filepaths, mosaic_class=MosaicGeometry, mosaic_kwargs=None, tile_kwargs=None,
                       stack_dimension='layer_id', tile_dimension='tile_id', **kwargs) -> "GeoTiffReader":
        """
        Creates a `GeoTiffReader` instance as one stack of GeoTIFF files.

        Parameters
        ----------
        filepaths : list of str
            List of full system paths to a GeoTIFF file.
        mosaic_class : geospade.raster.MosaicGeometry, optional
            Mosaic class used to manage the spatial properties of the file stack. If None, the most generic mosaic will
            be used by default. The initialised mosaic will only contain one tile.
        mosaic_kwargs : dict, optional
            Additional arguments for initialising `mosaic_class`.
        tile_kwargs : dict, optional
            Additional arguments for initialising a tile class associated with `mosaic_class`.
        stack_dimension : str, optional
            Dimension/column name of the dimension, where to stack the files along (first axis), e.g. time, bands etc.
            Defaults to 'layer_id', i.e. the layer ID's are used as the main coordinates to stack the files.
        tile_dimension : str, optional
            Dimension/column name of the dimension containing tile ID's in correspondence with the tiles in `mosaic`.
            Defaults to 'tile_id'.
        kwargs : dict, optional
            Key-word arguments for the `GeoTiffReader` constructor.

        Returns
        -------
        GeoTiffReader

        """
        mosaic_kwargs = mosaic_kwargs or dict()
        tile_kwargs = tile_kwargs or dict()

        n_filepaths = len(filepaths)
        file_register_dict = dict()
        file_register_dict['filepath'] = filepaths
        file_register_dict[tile_dimension] = ['0'] * n_filepaths
        file_register_dict[stack_dimension] = list(range(1, n_filepaths + 1))
        file_register = pd.DataFrame(file_register_dict)

        ref_filepath = filepaths[0]
        with GeoTiffFile(ref_filepath, 'r') as gt_file:
            sref_wkt = gt_file.sref_wkt
            geotrans = gt_file.geotrans
            n_rows, n_cols = gt_file.raster_shape

        tile_class = mosaic_class.get_tile_class()
        tile = tile_class(n_rows, n_cols, sref=SpatialRef(sref_wkt), geotrans=geotrans, name='0', **tile_kwargs)
        mosaic_geom = mosaic_class.from_tile_list([tile], check_consistency=False, **mosaic_kwargs)

        return cls(file_register, mosaic_geom, stack_dimension=stack_dimension, tile_dimension=tile_dimension,
                   **kwargs)

    @classmethod
    def from_mosaic_filepaths(cls, filepaths, mosaic_class=MosaicGeometry, mosaic_kwargs=None,
                              stack_dimension='layer_id', tile_dimension='tile_id', **kwargs) -> "GeoTiffReader":
        """
        Creates a `GeoTiffDataReader` instance as multiple stacks of GeoTIFF files.

        Parameters
        ----------
        filepaths : list of str
            List of full system paths to a GeoTIFF file.
        mosaic_class : geospade.raster.MosaicGeometry, optional
            Mosaic class used to manage the spatial properties of the file stacks. If None, the most generic mosaic will
            be used by default.
        mosaic_kwargs : dict, optional
            Additional arguments for initialising `mosaic_class`.
        stack_dimension : str, optional
            Dimension/column name of the dimension, where to stack the files along (first axis), e.g. time, bands etc.
            Defaults to 'layer_id', i.e. the layer ID's are used as the main coordinates to stack the files.
        tile_dimension : str, optional
            Dimension/column name of the dimension containing tile ID's in correspondence with the tiles in `mosaic`.
            Defaults to 'tile_id'.
        kwargs : dict, optional
            Key-word arguments for the `GeoTiffReader` constructor.

        Returns
        -------
        GeoTiffReader

        """
        mosaic_kwargs = mosaic_kwargs or dict()
        file_register_dict = dict()
        file_register_dict['filepath'] = filepaths
        tile_class = mosaic_class.get_tile_class()
        tiles, tile_ids, layer_ids = RasterDataReader._create_tile_and_layer_info_from_files(filepaths, tile_class,
                                                                                             GeoTiffFile)

        file_register_dict[tile_dimension] = tile_ids
        file_register_dict[stack_dimension] = layer_ids
        file_register = pd.DataFrame(file_register_dict)

        mosaic_geom = mosaic_class.from_tile_list(tiles, check_consistency=False, **mosaic_kwargs)

        return cls(file_register, mosaic_geom, stack_dimension=stack_dimension, tile_dimension=tile_dimension,
                   **kwargs)

    def apply_nan(self):
        """ Converts no data values given as an attribute '_FillValue' or keyword `nodatavals` to np.nan. """
        nodatavals = {dvar: self._ref_nodatavals[i] for i, dvar in enumerate(self.data_view.data_vars)}
        super().apply_nan(nodatavals=nodatavals)

    def read(self, bands=1, band_names=None, engine='vrt', n_cores=1,
             auto_decode=False, decoder=None, decoder_kwargs=None) -> "GeoTiffReader":
        """
        Reads data from disk.

        Parameters
        ----------
        bands : tuple of int or int, optional
            The GeoTIFF bands of interest. Defaults to the first band, i.e. 1.
        band_names : tuple of str or str, optional
            Names associated with the respective bands of the GeoTIFF files. Defaults to None, i.e. the band numbers
            will be used as a name.
        engine : str, optional
            Engine used in the background to read data. The following options are available:
                - 'vrt' : Uses GDAL's VRT format to stack the GeoTIFF files per tile and load them as once. This option
                          yields good performance if the mosaic is stored locally on one drive. Parallelisation is applied
                          across tiles.
                - 'parallel' : Reads file by file, but in a parallelised manner. This option yields good performance if
                               the mosaic is stored on a distributed file system.
        n_cores : int, optional
            Number of cores used to read data in a parallelised manner (defaults to 1).
        auto_decode : bool, optional
            True if data should be decoded according to the information available in its metadata. Defaults to False.
        decoder : callable, optional
            Function allowing to decode data read from disk.
        decoder_kwargs : dict, optional
            Keyword arguments for the decoder.

        """
        bands = to_list(bands)
        band_names = to_list(band_names)
        dst_tile = Tile.from_extent(self._mosaic.outer_extent, sref=self._mosaic.sref,
                                    x_pixel_size=self._mosaic.x_pixel_size,
                                    y_pixel_size=self._mosaic.y_pixel_size,
                                    name='0')

        shm_map = {band: self.__init_band_data(band, dst_tile) for band in bands}
        access_map = {src_tile.parent_root.name: GeoTiffAccess(src_tile, dst_tile) for src_tile in self._mosaic.tiles}
        data_mask = self.__create_data_mask_from(access_map, dst_tile.shape)

        if engine == 'vrt':
            self.__read_vrt_stack(access_map, shm_map, n_cores=n_cores, auto_decode=auto_decode, decoder=decoder,
                                  decoder_kwargs=decoder_kwargs)
        elif engine == 'parallel':
            self.__read_parallel(access_map, shm_map, n_cores=n_cores, auto_decode=auto_decode, decoder=decoder,
                                 decoder_kwargs=decoder_kwargs)
        else:
            err_msg = f"Engine '{engine}' is not supported!"
            raise ValueError(err_msg)

        data = {band: self.__load_band_data(band, shm_map, data_mask) for band in bands}

        self._data_geom = dst_tile
        self._data = self._to_xarray(data, band_names)
        self._add_grid_mapping()
        return self

    def __init_band_data(self, band, tile) -> Tuple[RawArray, tuple]:
        """
        Initialises shared memory array for a specific band.

        Parameters
        ----------
        band : int
            Band number.
        tile : Tile
            Tile containing information about the raster shape of the band data.

        Returns
        -------
        shm_ar : RawArray
            Shared memory array.
        shm_ar_shape : tuple
            Shape of the array.

        """
        np_dtype = np.dtype(self._ref_dtypes[band - 1])
        self._ref_nodatavals[band - 1] = np.array((self._ref_nodatavals[band - 1])).astype(np_dtype)
        data_nshm = np.ones((self.n_layers, tile.n_rows, tile.n_cols), dtype=np_dtype) * \
                    self._ref_nodatavals[band - 1]
        shm_ar_shape = data_nshm.shape
        c_dtype = np.ctypeslib.as_ctypes_type(data_nshm.dtype)
        shm_rar = RawArray(c_dtype, data_nshm.size)
        shm_data = np.frombuffer(shm_rar, dtype=np_dtype).reshape(shm_ar_shape)
        shm_data[:] = data_nshm[:]

        return shm_rar, shm_ar_shape

    def __load_band_data(self, band, shm_map, mask) -> np.ndarray:
        """
        Loads band data from the corresponding shared memory array.

        Parameters
        ----------
        band : int
            Band number.
        shm_map : dict
            Dictionary mapping the band number with a tuple containing a shared memory array and its shape.
        mask : np.array
            Data mask.

        Returns
        -------
        shm_ar : RawArray
            Shared memory array.

        """
        shm_rar, shm_ar_shape = shm_map[band]
        shm_data = np.frombuffer(shm_rar, dtype=self._ref_dtypes[band - 1]).reshape(shm_ar_shape)
        shm_data[:, ~mask.astype(bool)] = self._ref_nodatavals[band - 1]

        return shm_data

    def __create_data_mask_from(self, access_map, shape) -> np.ndarray:
        """
        Creates data mask from all individual tile masks contributing to the spatial extent of the data.

        Parameters
        ----------
        access_map : dict
            Dictionary mapping tile names with `GeoTiffAccess` instances for storing the access relation between
            the data mask and the tiles.
        shape : tuple
            Shape of the data mask.

        Returns
        -------
        data_mask : np.ndarray
            Data mask (1 valid data, 0 no data).

        """
        data_mask = np.ones(shape)
        for tile in self._mosaic.tiles:
            gt_access = access_map[tile.parent_root.name]
            data_mask[gt_access.dst_row_slice, gt_access.dst_col_slice] = tile.mask
        return data_mask

    def __read_vrt_stack(self, access_map, shm_map, n_cores=1,
                         auto_decode=False, decoder=None, decoder_kwargs=None):
        """
        Reads GeoTIFF data from a stack of GeoTIFF files by using GDAL's VRT format.

        Parameters
        ----------
        access_map : dict
            Dictionary mapping tile/geometry ID's with `GeoTiffAccess` instances to define the access patterns between
            the data to load and to assign.
        shm_map : dict
            Dictionary mapping band numbers with the respective name of the memory buffer of the shared numpy raw array.
        n_cores : int, optional
            Number of cores used to read data in a parallelised manner (defaults to 1).
        auto_decode : bool, optional
            True if data should be decoded according to the information available in its metadata. Defaults to False.
        decoder : callable, optional
            Function allowing to decode data read from disk.
        decoder_kwargs : dict, optional
            Keyword arguments for the decoder.

        """
        decoder_kwargs = decoder_kwargs or dict()
        global_file_register = self._file_register

        with Pool(n_cores, initializer=read_init, initargs=(global_file_register, access_map, shm_map,
                                                            self._file_dim, self._tile_dim,
                                                            auto_decode, decoder, decoder_kwargs)) as p:
            p.map(read_vrt_stack, access_map.keys())

    def __read_parallel(self, access_map, shm_map, n_cores=1,
                        auto_decode=False, decoder=None, decoder_kwargs=None):
        """
        Reads GeoTIFF mosaic on a file-by-file basis, in a parallel manner.

        Parameters
        ----------
        access_map : dict
            Dictionary mapping tile/geometry ID's with `GeoTiffAccess` instances to define the access patterns between
            the mosaic to load and to assign.
        shm_map : dict
            Dictionary mapping band numbers with the respective name of the memory buffer of the shared numpy raw array.
        n_cores : int, optional
            Number of cores used to read data in a parallelised manner (defaults to 1).
        auto_decode : bool, optional
            True if data should be decoded according to the information available in its metadata. Defaults to False.
        decoder : callable, optional
            Function allowing to decode data read from disk.
        decoder_kwargs : dict, optional
            Keyword arguments for the decoder.

        """
        decoder_kwargs = decoder_kwargs or None
        global_file_register = self._file_register

        with Pool(n_cores, initializer=read_init, initargs=(global_file_register, access_map, shm_map,
                                                            self._file_dim, self._tile_dim,
                                                            auto_decode, decoder, decoder_kwargs)) as p:
            p.map(read_single_files, global_file_register.index)

    def _to_xarray(self, data, band_names=None) -> xr.Dataset:
        """
        Converts data being available as a NumPy array to an xarray dataset.

        Parameters
        ----------
        data : dict
            Dictionary mapping band numbers to GeoTIFF raster data being available as a NumPy array.
        band_names : list of str, optional
            Band names associated with the respective band number.

        Returns
        -------
        xrds : xr.Dataset

        """
        dims = [self._file_dim] + self._ref_space_dims

        # retrieve file register with unique stack dimension
        ref_tile_id = list(self.file_register[self._tile_dim])[0]
        file_register_uni = self._file_register.loc[self._file_register[self._tile_dim] == ref_tile_id]

        coord_dict = dict()
        for coord in self._file_coords:
            coord_dict[coord] = file_register_uni[coord]

        coord_dict[self._ref_space_dims[0]] = self._data_geom.y_coords
        coord_dict[self._ref_space_dims[1]] = self._data_geom.x_coords

        xar_dict = dict()
        bands = list(data.keys())
        band_names = band_names or bands
        for i, band in enumerate(bands):
            xar_dict[band_names[i]] = xr.DataArray(data[band], coords=coord_dict, dims=dims,
                                                   attrs={'_FillValue': self._ref_nodatavals[band - 1]})

        xrds = xr.Dataset(data_vars=xar_dict)

        return xrds


class GeoTiffWriter(RasterDataWriter):
    """ Allows to write and manage a stack of GeoTIFF files. """
    def __init__(self, mosaic, file_register=None, data=None, stack_dimension='layer_id', stack_coords=None,
                 tile_dimension='tile_id', dirpath=None, fn_pattern='{layer_id}.tif', fn_formatter=None):
        """
        Constructor of `GeoTiffWriter`.

        Parameters
        ----------
        mosaic : geospade.raster.MosaicGeometry
            Mosaic representing the spatial allocation of the given files. The tiles of the mosaic have to match the
            ID's/names of the 'tile_id' column.
        file_register : pd.Dataframe, optional
            Data frame managing a stack/list of files containing the following columns:
                - 'filepath' : str
                    Full file path to a geospatial file.
                - 'layer_id' : object
                    Specifies an ID to which layer a file belongs to, e.g. a layer counter or a timestamp. Must
                    correspond to `stack_dimension`.
                - 'tile_id' : str
                    Tile name or ID to which tile a file belongs to.
            If it is None, then the constructor tries to create a file from other keyword arguments, i.e. `data`,
            `dirpath`, `fn_pattern`, and `fn_formatter`.
        data : xr.Dataset, optional
            Raster data stored in memory. It must match the spatial sampling and CRS of the mosaic, but not its spatial
            extent or tiling. Moreover, the dimension of the mosaic along the first dimension (stack dimension), must
            match the entries/filepaths in `file_register`.
        stack_dimension : str, optional
            Dimension/column name of the dimension, where to stack the files along (first axis), e.g. time, bands etc.
            Defaults to 'layer_id', i.e. the layer ID's are used as the main coordinates to stack the files.
        stack_coords : list, optional
            Additional columns of `file_register` to use as coordinates. Defaults to None, i.e. only coordinates along
            `stack_dimension` are used.
        tile_dimension : str, optional
            Dimension/column name of the dimension containing tile ID's in correspondence with the tiles in `mosaic`.
            Defaults to 'tile_id'.
        dirpath : str, optional
            Directory path to the folder where the GeoTIFF files should be written to. Defaults to None, i.e. the
            current working directory is used.
        fn_pattern : str, optional
            Pattern for the filename of the new GeoTIFF files. To fill in specific parts of the new file name with
            information from the file register, you can specify the respective file register column names in curly
            brackets and add them to the pattern string as desired. Defaults to '{layer_id}.tif'.
        fn_formatter : dict, optional
            Dictionary mapping file register column names with functions allowing to encode their values as strings.

        """

        super().__init__(mosaic, file_register=file_register, data=data, stack_dimension=stack_dimension,
                         stack_coords=stack_coords, tile_dimension=tile_dimension, dirpath=dirpath,
                         fn_pattern=fn_pattern, fn_formatter=fn_formatter)

    def __get_encoding_info_from_data(self, data, band_names) -> Tuple[dict, dict, dict, dict]:
        """
        Extracts encoding information from an xarray dataset.

        Parameters
        ----------
        data : xr.Dataset
            Data to extract encoding info from.
        band_names : list of str
            List of band names.

        Returns
        -------
        nodatavals : dict
            Band number mapped to no data value (defaults to 0).
        scale_factors : dict
            Band number mapped to scale factor (defaults to 1).
        offsets : dict
            Band number mapped to offset (defaults to 0).
        dtypes : dict
            Band number mapped to data type.

        """
        nodatavals = dict()
        scale_factors = dict()
        offsets = dict()
        dtypes = dict()
        for i, band_name in enumerate(band_names):
            dtypes[i + 1] = data[band_name].data.dtype.name
            nodatavals[i + 1] = data[band_name].attrs.get('_FillValue', 0)
            scale_factors[i + 1] = data[band_name].attrs.get('scale_factor', 1)
            offsets[i + 1] = data[band_name].attrs.get('add_offset', 0)

        return nodatavals, scale_factors, offsets, dtypes

    def write(self, data, use_mosaic=False, data_variables=None, encoder=None, encoder_kwargs=None, overwrite=False,
              **kwargs):
        """
        Writes a certain chunk of data to disk.

        Parameters
        ----------
        data : xr.Dataset
            Data chunk to be written to disk or being appended to existing data.
        use_mosaic : bool, optional
            True if data should be written according to the mosaic.
            False if data composes a new tile and should not be tiled (default).
        data_variables : list of str, optional
            Data variables to write. Defaults to None, i.e. all data variables are written.
        encoder : callable, optional
            Function allowing to encode data before writing it to disk.
        encoder_kwargs : dict, optional
            Keyword arguments for the encoder.
        overwrite : bool, optional
            True if data should be overwritten, False if not (default).

        """
        data_geom = self.raster_geom_from_data(data, sref=self.mosaic.sref)
        data_filt = data if data_variables is None else data[data_variables]
        band_names = list(data_filt.data_vars)
        n_bands = len(band_names)
        nodatavals, scale_factors, offsets, dtypes = self.__get_encoding_info_from_data(data_filt, band_names)

        for filepath, file_group in self._file_register.groupby('filepath'):
            tile_id = file_group.iloc[0].get(self._tile_dim, '0')

            file_coords = list(file_group[self._file_dim])
            xrds = data_filt.sel(**{self._file_dim: file_coords})
            data_write = xrds[band_names].to_array().data

            if use_mosaic:
                src_tile = self._mosaic[tile_id]
                if not src_tile.intersects(data_geom):
                    continue
                dst_tile = data_geom.slice_by_geom(src_tile, inplace=False, name='0')
                gt_access = GeoTiffAccess(dst_tile, src_tile, src_root_raster_geom=data_geom)
                data_write = data_write[..., gt_access.src_row_slice, gt_access.src_col_slice]
            else:
                src_tile = data_geom
                dst_tile = data_geom
                gt_access = GeoTiffAccess(dst_tile, src_tile, src_root_raster_geom=data_geom)

            file_id = file_group.iloc[0].get('file_id', None)
            if file_id is None:
                gt_file = GeoTiffFile(filepath, mode='w', geotrans=src_tile.geotrans, sref_wkt=src_tile.sref.wkt,
                                      raster_shape=src_tile.shape, n_bands=n_bands, dtypes=dtypes,
                                      scale_factors=scale_factors, offsets=offsets, nodatavals=nodatavals)
                file_id = len(list(self._files.keys())) + 1
                self._files[file_id] = gt_file
                self._file_register.loc[file_group.index, 'file_id'] = file_id

            gt_file = self._files[file_id]
            gt_file.write(data_write[..., gt_access.src_row_slice,
                                     gt_access.src_col_slice].reshape((-1, dst_tile.n_rows, dst_tile.n_cols)),
                          row=gt_access.dst_window[0], col=gt_access.dst_window[1],
                          encoder=encoder, encoder_kwargs=encoder_kwargs)

    def export(self, use_mosaic=False, data_variables=None, encoder=None, encoder_kwargs=None, overwrite=False,
               **kwargs):
        """
        Writes all internally stored data to disk.

        Parameters
        ----------
        use_mosaic : bool, optional
            True if data should be written according to the mosaic.
            False if data composes a new tile and should not be tiled (default).
        data_variables : list of str, optional
            Data variables to write. Defaults to None, i.e. all data variables are written.
        encoder : callable, optional
            Function allowing to encode data before writing it to disk.
        encoder_kwargs : dict, optional
            Keyword arguments for the encoder.
        overwrite : bool, optional
            True if data should be overwritten, False if not (default).

        """

        self.write(self.data_view, use_mosaic, data_variables, encoder, encoder_kwargs, overwrite, **kwargs)


def read_vrt_stack(tile_id):
    """
    Function being responsible to create a new VRT file from a stack of GeoTIFF files, read data from this files, and
    assign it to a shared memory array. This function is meant to be executed in parallel on different cores.

    Parameters
    ----------
    tile_id : str
        Tile/geometry ID coming from the Pool's mapping function.

    """
    global_file_register = PROC_OBJS['global_file_register']
    access_map = PROC_OBJS['access_map']
    shm_map = PROC_OBJS['shm_map']
    tile_dimension = PROC_OBJS['tile_dimension']

    gt_access = access_map[tile_id]
    bands = list(shm_map.keys())
    file_register = global_file_register.loc[global_file_register[tile_dimension] == tile_id]

    path = tempfile.gettempdir()
    vrt_filepath = os.path.join(path, f"{datetime.utcnow().strftime('%Y%m%d%H%M%S')}-{uuid.uuid4().hex}.vrt")
    while os.path.exists(vrt_filepath):
        vrt_filepath = os.path.join(path, f"{datetime.utcnow().strftime('%Y%m%d%H%M%S')}-{uuid.uuid4().hex}.vrt")
    filepaths = list(file_register['filepath'])
    create_vrt_file(filepaths, vrt_filepath, gt_access.src_shape, gt_access.src_wkt, gt_access.src_geotrans,
                    bands=bands)

    src = gdal.Open(vrt_filepath, gdal.GA_ReadOnly)
    vrt_data = src.ReadAsArray(*gt_access.gdal_args)
    for band in bands:
        _assign_vrt_stack_per_band(tile_id, band, src, vrt_data)


def _assign_vrt_stack_per_band(tile_id, band, src, vrt_data):
    """
    Assigns loaded raster data to shared memory array for a specific band.

    Parameters
    ----------
    tile_id : str
        Tile/geometry ID coming from the Pool's mapping function.
    band : int
        Band number.
    src : gdal.Dataset
        GDAL dataset handle.
    vrt_data : np.ndarray
        In-memory data read from disk.

    """
    auto_decode = PROC_OBJS['auto_decode']
    decoder = PROC_OBJS['decoder']
    decoder_kwargs = PROC_OBJS['decoder_kwargs']
    access_map = PROC_OBJS['access_map']
    shm_map = PROC_OBJS['shm_map']

    gt_access = access_map[tile_id]
    bands = list(shm_map.keys())
    n_bands = len(bands)
    b_idx = bands.index(band)

    band_data = vrt_data[b_idx::n_bands, ...]
    scale_factor = src.GetRasterBand(b_idx + 1).GetScale()
    nodataval = src.GetRasterBand(b_idx + 1).GetNoDataValue()
    offset = src.GetRasterBand(b_idx + 1).GetOffset()
    dtype = GDAL_TO_NUMPY_DTYPE[src.GetRasterBand(b_idx + 1).DataType]
    if auto_decode:
        band_data = band_data.astype(float)
        band_data[band_data == nodataval] = np.nan
        band_data = band_data * scale_factor + offset
    else:
        if decoder is not None:
            band_data = decoder(band_data, nodataval=nodataval, band=band, scale_factor=scale_factor,
                                offset=offset,
                                dtype=dtype, **decoder_kwargs)

    shm_rar, shm_ar_shape = shm_map[band]
    shm_data = np.frombuffer(shm_rar, dtype=dtype).reshape(shm_ar_shape)
    shm_data[:, gt_access.dst_row_slice, gt_access.dst_col_slice] = band_data


def read_single_files(file_idx):
    """
    Function being responsible to read data from a single GeoTIFF file and assign it to a shared memory array.
    This function is meant to be executed in parallel on different cores.

    Parameters
    ----------
    file_idx : any
        Index value to access a specific row of the file register. The actual value should come from the Pool's
        mapping function.

    """
    global_file_register = PROC_OBJS['global_file_register']
    access_map = PROC_OBJS['access_map']
    shm_map = PROC_OBJS['shm_map']
    auto_decode = PROC_OBJS['auto_decode']
    decoder = PROC_OBJS['decoder']
    decoder_kwargs = PROC_OBJS['decoder_kwargs']
    stack_dimension = PROC_OBJS['stack_dimension']
    tile_dimension = PROC_OBJS['tile_dimension']

    file_entry = global_file_register.loc[file_idx]
    layer_id = file_entry[stack_dimension]
    tile_id = file_entry[tile_dimension]
    filepath = file_entry['filepath']
    gt_access = access_map[tile_id]
    bands = list(shm_map.keys())

    with GeoTiffFile(filepath, mode='r', auto_decode=auto_decode) as gt_file:
        gt_data = gt_file.read(*gt_access.read_args, bands=bands, decoder=decoder, decoder_kwargs=decoder_kwargs)
        for band in bands:
            dtype = gt_file.dtypes[band]
            shm_rar, shm_ar_shape = shm_map[band]
            shm_data = np.frombuffer(shm_rar, dtype=dtype).reshape(shm_ar_shape)
            shm_data[layer_id, gt_access.dst_row_slice, gt_access.dst_col_slice] = gt_data[(band - 1), ...]


if __name__ == '__main__':
    pass



from netcdf_common import *
from veranda.raster.native.netcdf import NetCdfXrFile


def test_read_write(filepath, three_var_ds):

    with NetCdfXrFile(filepath, mode='w', data_variables=['sig', 'inc', 'azi'],
                      nodatavals={'inc': -9999, 'azi': -9999}, dtypes={'inc': 'int32', 'azi': 'int32'}) as nc:
        nc.write(three_var_ds)

    with NetCdfXrFile(filepath) as nc:
        ds = nc.read()
        np.testing.assert_array_equal(ds['sig'][:], three_var_ds['sig'][:])
        np.testing.assert_array_equal(ds['inc'][:], three_var_ds['inc'][:])
        np.testing.assert_array_equal(ds['azi'][:], three_var_ds['azi'][:])


def test_use_auto_decoding(filepath, complex_three_var_ds):
    with NetCdfXrFile(filepath, mode='w') as nc:
        nc.write(complex_three_var_ds)

    with NetCdfXrFile(filepath, mode='r', auto_decode=True) as nc:
        ds = nc.read()
        np.testing.assert_array_equal(ds['sig'][:], complex_three_var_ds['sig'][:] * 2 + 3)
        np.testing.assert_array_equal(ds['inc'][:], complex_three_var_ds['inc'][:] * 2)
        np.testing.assert_array_equal(ds['azi'][:], complex_three_var_ds['azi'][:])


def test_ignore_auto_decoding(filepath, complex_three_var_ds):
    with NetCdfXrFile(filepath, mode='w') as nc:
        nc.write(complex_three_var_ds)

    with NetCdfXrFile(filepath, mode='r', auto_decode=False) as nc:
        ds = nc.read()
        np.testing.assert_array_equal(ds['sig'][:], complex_three_var_ds['sig'][:])
        np.testing.assert_array_equal(ds['inc'][:], complex_three_var_ds['inc'][:])
        np.testing.assert_array_equal(ds['azi'][:], complex_three_var_ds['azi'][:])


def test_chunksizes(filepath, simple_ds):
    chunksizes = (100, 10, 10)
    with NetCdfXrFile(filepath, mode='w', data_variables=list(simple_ds.data_vars), chunksizes=chunksizes) as nc:
        nc.write(simple_ds)

    with NetCdfXrFile(filepath, mode='r') as nc:
        ds = nc.read()
        assert ds['sig'].data.chunksize == chunksizes
        assert ds['inc'].data.chunksize == chunksizes


def test_time_units(filepath, simple_ds):
    time_units = 'days since 2000-01-01 00:00:00'
    with NetCdfXrFile(filepath, mode='w', attrs={'time': {'units': time_units}}) as nc:
        nc.write(simple_ds)

    with NetCdfXrFile(filepath) as nc:
        ds = nc.read()
        np.testing.assert_array_equal(pd.DatetimeIndex(ds['time'].data),
                                      simple_ds.coords['time'])


def test_geotransform(filepath, simple_ds):
    xdim = 100
    ydim = 100
    geotrans = (3000000.0, 500.0, 0.0, 1800000.0, 0.0, -500.0)
    with NetCdfXrFile(filepath, mode='w', geotrans=geotrans) as nc:
        nc.write(simple_ds)

    with NetCdfXrFile(filepath) as nc:
        ds = nc.read()

    x = geotrans[0] + (0.5 + np.arange(xdim)) * geotrans[1] + \
        (0.5 + np.arange(xdim)) * geotrans[2]
    y = geotrans[3] + (0.5 + np.arange(ydim)) * geotrans[4] + \
        (0.5 + np.arange(ydim)) * geotrans[5]

    np.testing.assert_array_equal(ds['x'].values, x)
    np.testing.assert_array_equal(ds['y'].values, y)


def test_non_temporal_read_and_write(filepath, nt_ds):
    with NetCdfXrFile(filepath, mode='w', data_variables=['sig', 'inc', 'azi'], stack_dims={'layer': None},
                      nodatavals={'inc': -9999, 'azi': -9999}, dtypes={'inc': 'int32', 'azi': 'int32'}) as nc:
        nc.write(nt_ds)

    with NetCdfXrFile(filepath) as nc:
        ds = nc.read()
        np.testing.assert_array_equal(ds['sig'][:], nt_ds['sig'][:])
        np.testing.assert_array_equal(ds['inc'][:], nt_ds['inc'][:])
        np.testing.assert_array_equal(ds['azi'][:], nt_ds['azi'][:])



import os
import pytest
import numpy as np
import pandas as pd
import xarray as xr
from tempfile import mkdtemp


@pytest.fixture
def filepath():
    return os.path.join(mkdtemp(), 'test.nc')


@pytest.fixture
def simple_ds():
    data = np.ones((100, 100, 100), dtype=np.float32)
    dims = ['time', 'y', 'x']
    coords = {'time': pd.date_range('2000-01-01', periods=data.shape[0])}

    ds = xr.Dataset({'sig': (dims, data),
                     'inc': (dims, data)}, coords=coords)

    return ds


@pytest.fixture
def three_var_ds():
    data = np.ones((100, 100, 100), dtype=np.float32)
    dims = ['time', 'y', 'x']
    coords = {'time': pd.date_range('2000-01-01', periods=data.shape[0])}
    attr1 = {'unit': 'dB'}
    attr2 = {'unit': 'degree'}

    ds = xr.Dataset({'sig': (dims, data, attr1),
                     'inc': (dims, data, attr2),
                     'azi': (dims, data, attr2)}, coords=coords)
    return ds


@pytest.fixture
def complex_three_var_ds():
    data = np.ones((100, 100, 100), dtype=np.float32)
    dims = ['time', 'y', 'x']
    coords = {'time': pd.date_range('2000-01-01', periods=data.shape[0])}
    attr1 = {'unit': 'dB', 'scale_factor': 2, 'add_offset': 3, 'fill_value': -9999}
    attr2 = {'unit': 'degree', '_FillValue': -9999, 'scale_factor': 2, 'add_offset': 0}
    attr3 = {'unit': 'degree', '_FillValue': -9999}

    ds = xr.Dataset({'sig': (dims, data, attr1),
                     'inc': (dims, data, attr2),
                     'azi': (dims, data, attr3)}, coords=coords)

    return ds


@pytest.fixture
def nt_ds():
    data = np.ones((100, 100, 100), dtype=np.float32)
    dims = ['layer', 'y', 'x']
    coords = {'layer': range(data.shape[0])}
    attr1 = {'unit': 'dB'}
    attr2 = {'unit': 'degree'}

    ds = xr.Dataset({'sig': (dims, data, attr1),
                     'inc': (dims, data, attr2),
                     'azi': (dims, data, attr2)}, coords=coords)

    return ds
